
/*
 * +==========================================================
 * |	'id_ca.c':
 * |	Resource cache manager.
 * |
 * |	LoneWolf (Wolf3D/Spear of Destiny engine) source code
 * |    by D.G.Gayev (dgaev@mail.ru), 2004-2005
 * +==========================================================
 *
 */

#include "wl_def.h"

/*
=============================================================================

						 GLOBAL VARIABLES

=============================================================================
*/

word RLEWtag;

word	*mapsegs[MAPPLANES];
maptype	*mapheaderseg[NUMMAPS];
static byte	**audiosegs;
static byte	**grsegs;

char extension[5];

// Read-only resources:

// (graphics)
static char gheadname[] = "vgahead";
static char gfilename[] = "vgagraph";
static char gdictname[] = "vgadict";

// (maps)
static char mheadname[] = "maphead";
static char mfilename[] = "gamemaps";

// (audio)
static char aheadname[] = "audiohed";
static char afilename[] = "audiot";

// (page file)
static char pfilename[] = "vswap";

static long *grstarts;		/* array of offsets in "vgagraph" */
static long *audiostarts;	/* array of offsets in "audiot" */

typedef struct {	/* 0-255 is a character, > is a pointer to a node */
	int bit0, bit1;
} huffnode;

static huffnode grhuffman[256];

static int grhandle = -1;	/* handle to VGAGRAPH */
static int maphandle = -1;	/* handle to GAMEMAPS */
static int audiohandle = -1;	/* handle to AUDIOT */

enum { NO_ENTRY = 0xffff };

extern word GFX_RSM [];
extern word SND_RSM [];
extern word MUS_RSM [];

// (Debug messages output level)
#define	CA_DL	1

// Locate game resource
char *resource_name (const char *base_name, boolean global) {
static char fname[256];
char *path = global ? path_resource : path_user;
char *ext = extension;

if (path) strcpy (fname, path);
else *fname = '\0';

//
// (perform name translation)
//

if (base_name == aheadname || base_name == afilename ||
    base_name == gdictname || base_name == gheadname || base_name == gfilename)
	if (is_game == GAME_SD2 || is_game == GAME_SD3)
		ext = "sod";

strcat(fname, base_name);
strcat(fname, ".");
strcat(fname, ext);
return fname;
}	// resource_name

/*
=============================================================================

					   LOW LEVEL ROUTINES

=============================================================================
*/

int TryOpenRead (char *name) {
int handle = OpenRead(name);
if (handle == -1) {
	Quit("Can't open file \"%s\"!", name);
	}
return handle;
}

/*
==========================
=
= CA_WriteFile
=
= Writes a file from a memory buffer
=
==========================
*/

boolean CA_WriteFile(char *filename, void *ptr, long length)
{
	ssize_t l;
	int handle;

	handle = OpenWrite(filename);

	if (handle == -1)
		return false;

	l = WriteBytes(handle, (byte *)ptr, length);
	if (l == -1) {
		perror("CA_FarWrite");
		return false;
	} else if (l == 0) {
		fprintf(stderr, "CA_FarWrite hit EOF?\n");
		return false;
	} else if (l != length) {
		fprintf(stderr, "CA_FarWrite only wrote %ld out of %ld\n", l, length);
		return false;
	}

	CloseWrite(handle);
	
	return true;
}

/*
==========================
=
= CA_LoadFile
=
= Allocate space for and load a file
=
==========================
*/

boolean CA_LoadFile(char *filename, byte **ptr)
{
	int handle;
	ssize_t l;
	long size;

	if ((handle = OpenRead(filename)) == -1)
		return false;

	size = ReadLength(handle);
	(*ptr) = MM_Alloc(size, "filedata");
	
	l = ReadBytes(handle, (*ptr), size);
	
	if (l == -1) {
		perror("CA_FarRead");
		return false;
	} else if (l == 0) { 
		fprintf(stderr, "CA_FarRead hit EOF?\n");
		return false;
	} else if (l != size) {
		fprintf(stderr, "CA_FarRead only read %ld out of %ld\n", l, size);
		return false;
	}
	
	CloseRead(handle);
	
	return true;
}

/*
============================================================================

		COMPRESSION routines

============================================================================
*/

/*
======================
=
= CAL_HuffExpand
= Length is the length of the EXPANDED data
=
======================
*/

/* From Ryan C. Gordon -- ryan_gordon@hotmail.com */
void CAL_HuffExpand(byte *source, byte *dest, long length, huffnode *htable)
{
	huffnode *headptr;
	huffnode *nodeon;
	byte      mask = 1;
	word      path;
	byte     *endoff = dest + length;
	byte val = *source++;

	nodeon = headptr = htable + 254;

	do {
			if ((path = (val & mask) ? nodeon->bit1 : nodeon->bit0) < 256) {
			*dest++ = (byte)path;
			nodeon = headptr;
		} else
			nodeon = htable + (path - 256);

      if ((mask <<= 1) == 0) {
				mask = 1;
				val = *source++;
				}
	} while (dest != endoff);
}	// CAL_HuffExpand

/*
======================
=
= CAL_CarmackExpand
= Length is the length of the EXPANDED data
=
= Note: This function happens to implicity swap words' bytes around.
=       For maps, this happens to be the desired effect.
=
======================
*/

#define NEARTAG	0xa7
#define FARTAG	0xa8

void CAL_CarmackExpand(byte *source, word *dest, word length)
{
	unsigned int offset;
	word *copyptr, *outptr;	
	byte chhigh, chlow, *inptr;
	
	length /= 2;

	inptr = source;
	outptr = dest;

	while (length) {		
		chlow = *inptr++; /* count */
		chhigh = *inptr++;
		
		if (chhigh == NEARTAG) {
			if (!chlow) {	
				/* have to insert a word containing the tag byte */
				*outptr++ = (chhigh << 8) | *inptr++;
				length--;
			} else {
				offset = *inptr++;
				
				copyptr = outptr - offset;
				
				length -= chlow;
				while (chlow--)
					*outptr++ = *copyptr++;
			}
		} else if (chhigh == FARTAG) {
			if (!chlow) {
				/* have to insert a word containing the tag byte */
				*outptr++ = (chhigh << 8) | *inptr++;
				length--;
			} else {
				offset = *inptr | (inptr[1] << 8);
				inptr += 2;

				copyptr = dest + offset;
				length -= chlow;
				while (chlow--)
					*outptr++ = *copyptr++;
			}
		} else {
			*outptr++ = (chhigh << 8) | chlow;
			length--;
		}
	}
}

/*
======================
=
= CA_RLEWexpand
= length is EXPANDED length
=
======================
*/

void CA_RLEWexpand(word *source, word *dest, long length, word rlewtag)
{
	word value, count;
	word *end = dest + length / 2;

	/* expand it */
	do {	if ((value = *source++) != rlewtag)
			/* uncompressed */
			*dest++ = value;
		else {
			/* compressed string */
			count = *source++;
			value = *source++;
			while (count--) *dest++ = value;
		}
	} while (dest < end);
}	// CA_RLEWexpand

/*
=============================================================================

					 CACHE MANAGER ROUTINES

=============================================================================
*/

/*
======================
=
= CAL_SetupGrFile
=
======================
*/

static pictabletype *pictable;

static void CAL_SetupGrFile()
{
	int handle;
	byte *grtemp;
	int i;

	grsegs = MM_Alloc (sizeof (byte *) * NUMGRCHUNKS, "grsegs");
	for (i = 0; i != NUMGRCHUNKS; ++ i)
		grsegs[i] = 0;

	/* load huffman dictionary for graphics files */
	handle = TryOpenRead(resource_name(gdictname, true));

	for (i = 0; i != 256; i++) {
		grhuffman[i].bit0 = ReadInt16(handle);
		grhuffman[i].bit1 = ReadInt16(handle);
		}

	CloseRead(handle);
	
	/* load the data offsets from vgahead.ext */
	grstarts = MM_Alloc((NUMGRCHUNKS+1)*4, "grstarts");
	grtemp = MM_Alloc((NUMGRCHUNKS+1)*3, 0);

	handle = TryOpenRead(resource_name(gheadname, true));

	ReadBytes(handle, grtemp, (NUMGRCHUNKS+1)*3);

	for (i = 0; i < NUMGRCHUNKS+1; i++)
		grstarts[i] =
		(grtemp[i*3+0]<<0)|(grtemp[i*3+1]<<8)|(grtemp[i*3+2]<<16);

	MM_Free(grtemp);

	CloseRead(handle);
	
	/* Open the graphics file, leaving it open until the game is finished */

	grhandle = TryOpenRead(resource_name(gfilename, true));

	/* load the pic headers into pictable */
	CA_CacheGrChunk(STRUCTPIC);
	
	pictable = MM_Alloc (sizeof(pictabletype) * NUMPICS, "pictable");

	grtemp = CA_GetGrChunk(STRUCTPIC);
	for (i = 0; i != NUMPICS; i ++) {
		pictable[i].width = grtemp[i*4+0] | (grtemp[i*4+1] << 8);
		pictable[i].height = grtemp[i*4+2] | (grtemp[i*4+3] << 8);
		}

	CA_UnCacheGrChunk(STRUCTPIC);

Diag(0, "Graphics files init: %u total (%u fonts & tiles, %u pics, %u externs)",
		NUMGRCHUNKS, 4, NUMPICS, NUMEXTERNS);
}	// CAL_SetupGrFile

void GetPicDims (int ichunk, int *pwidth, int *pheight) {
// (NOTE: pictable is ordered by phisical pics order)
pictabletype *entry =
	&pictable [GFX_RSM [ichunk-GFX_BEG] - (STARTPICS-GFX_BEG)];
if (pwidth) *pwidth = entry->width;
if (pheight) *pheight = entry->height;
}	// GetPicDims

/* ======================================================================== */

/*
======================
=
= CAL_SetupMapFile
=
======================
*/

static void CAL_SetupMapFile()
{
	int i;
	int handle;
	long pos;
  unsigned count = 0;

	handle = TryOpenRead(resource_name(mheadname, true));

	RLEWtag = ReadInt16(handle);

/* open the data file */

	maphandle = TryOpenRead(resource_name(mfilename, true));

/* load all map header */
	for (i = 0; i != NUMMAPS; i++)
	{
		pos = ReadInt32(handle);
		if (pos == 0) {
			mapheaderseg[i] = NULL;
			continue;
		}

    ++ count;

		mapheaderseg[i] = MM_Alloc(sizeof(maptype), "mapheader");

		ReadSeek(maphandle, pos, SEEK_SET);

		mapheaderseg[i]->planestart[0] = ReadInt32(maphandle);
		mapheaderseg[i]->planestart[1] = ReadInt32(maphandle);
		mapheaderseg[i]->planestart[2] = ReadInt32(maphandle);

		mapheaderseg[i]->planelength[0] = ReadInt16(maphandle);
		mapheaderseg[i]->planelength[1] = ReadInt16(maphandle);
		mapheaderseg[i]->planelength[2] = ReadInt16(maphandle);
		mapheaderseg[i]->width = ReadInt16(maphandle);
		mapheaderseg[i]->height = ReadInt16(maphandle);
		ReadBytes(maphandle, (byte *)mapheaderseg[i]->name, 16);
	}

	CloseRead(handle);

/* allocate space for MAPPLANES MAPHGT*MAPWDT planes */
	for (i = 0; i != MAPPLANES; i++) {
		mapsegs[i] = MM_Alloc (MAPWDT*MAPHGT*2, "mapseg");
	}

Diag(0, "Map files init: %u total [%u width * %u height * %u planes]",
		count, MAPWDT, MAPHGT, MAPPLANES);
}

/* ======================================================================== */

/*
======================
=
= CAL_SetupAudioFile
=
======================
*/

static void CAL_SetupAudioFile()
{
	int handle;
	long length;
	int i;

	audiosegs = MM_Alloc(sizeof (byte *) * NUMSNDCHUNKS, "audiosegs");
	for (i = 0; i != NUMSNDCHUNKS; ++ i)
		audiosegs[i] = 0;

	handle = TryOpenRead(resource_name(aheadname, true));

	length = ReadLength(handle);

	audiostarts = MM_Alloc (length, "audiostarts");

	length /= 4;
	for (i = 0; i != length; i ++)
		audiostarts[i] = ReadInt32(handle);

	CloseRead(handle);	

/* open the data file */

	audiohandle = TryOpenRead(resource_name(afilename, true));

Diag(0, "Audio files init: %u total (%u FM sounds, %u music)",
		NUMSNDCHUNKS, NUMSOUNDS, NUMSNDCHUNKS - STARTMUSIC);
}

/* ======================================================================== */

/*
======================
=
= CA_Startup
=
= Open all files and load in headers
=
======================
*/

void CA_Startup()
{
	CAL_SetupMapFile();
	CAL_SetupGrFile();
	CAL_SetupAudioFile();
}

/*
======================
=
= CA_Shutdown
=
= Closes all files
=
======================
*/

void CAL_ShutdownGrFile() {
  unsigned i;

	for (i = 0; i != NUMGRCHUNKS; ++ i)
		MM_Free (grsegs[i]);

	MM_Free (grsegs);
	MM_Free (grstarts);
	MM_Free (pictable);

	CloseRead(grhandle);

Diag(0, "Graphics files shutdown OK");
}

void CAL_ShutdownAudioFile() {
  unsigned i;

	for (i = 0; i != NUMSNDCHUNKS; ++ i)
		MM_Free (audiosegs[i]);

  MM_Free (audiosegs);
  MM_Free (audiostarts);

	CloseRead(audiohandle);

Diag(0, "Audio files shutdown OK");
}

void CAL_ShutdownMapFile() {
  unsigned i;

	for (i = 0; i != NUMMAPS; i++)
		MM_Free (mapheaderseg[i]);

	for (i = 0; i != MAPPLANES; i++)
		MM_Free (mapsegs[i]);

	CloseRead(maphandle);

Diag(0, "Map files shutdown OK");
}

void CA_Shutdown()
{
CAL_ShutdownGrFile();
CAL_ShutdownAudioFile();
CAL_ShutdownMapFile();
}

/* ======================================================================== */

/*
======================
=
= CA_CacheAudioChunk
=
======================
*/

void CA_CacheAudioChunk(int chunk)
{
	int pos, length, pchunk;

	// (translate chunk #)

	if (SND_BEG <= chunk && chunk < SND_END) {
	    pchunk = SND_RSM [chunk - SND_BEG];
	    if (pchunk == NO_ENTRY)
		Quit ("CA_CacheAudioChunk: undefined chunk #%d!", chunk);
	    }
	else Quit ("CA_CacheAudioChunk: illegal chunk #%d!", chunk);
	pchunk += STARTADLIBSOUNDS;

	Diag(CA_DL, "CA_CacheAudioChunk %d (%d)", chunk, pchunk);

	if (audiosegs[pchunk])
		return;

	pos = audiostarts[pchunk];
	length = audiostarts[pchunk+1]-pos;

	ReadSeek(audiohandle, pos, SEEK_SET);

	audiosegs[pchunk] = MM_Alloc(length, "audiochunk");

	ReadBytes(audiohandle, audiosegs[pchunk], length);
}	// CA_CacheAudioChunk

void CA_UnCacheAudioChunk(int chunk)
{	int pchunk;

	// (translate chunk #)

	if (SND_BEG <= chunk && chunk < SND_END) {
	    pchunk = SND_RSM [chunk - SND_BEG];
	    if (pchunk == NO_ENTRY)
		Quit ("CA_UnCacheAudioChunk: undefined chunk #%d!", chunk);
	    }
	else Quit ("CA_UnCacheAudioChunk: illegal chunk #%d!", chunk);
	pchunk += STARTADLIBSOUNDS;

	Diag(CA_DL, "CA_UnCacheAudioChunk %d (%d)", chunk, pchunk);

	if (audiosegs[pchunk] == 0) {
		Diag(0, "Trying to free null audio chunk %d!", pchunk);
		return;
	}

	MM_Free(audiosegs[pchunk]);
	audiosegs[pchunk] = NULL;
}	// CA_UnCacheAudioChunk

byte *CA_GetAudioChunk (int chunk) {

int pchunk;

// (translate chunk #)

if (SND_BEG <= chunk && chunk < SND_END)
	pchunk = SND_RSM [chunk - SND_BEG];
else Quit ("CA_GetAudioChunk: illegal chunk #%d!", chunk);

if (pchunk == NO_ENTRY) Quit ("CA_GetAudioChunk: unset chunk #%d!", chunk);

Diag(CA_DL, "CA_GetAudioChunk %d (%d)", chunk, pchunk);

if (pchunk < NUMSOUNDS) return audiosegs[STARTADLIBSOUNDS+pchunk];

Quit("CA_GetAudioChunk: unknown pchunk #%d", pchunk);
return 0;
}	// CA_GetAudioChunk

void CA_CacheMusicChunk(int chunk)
{
	int pos, length, pchunk;

	// (translate chunk #)

	if (MUS_BEG <= chunk && chunk < MUS_END) {
	    pchunk = MUS_RSM [chunk - MUS_BEG];
	    if (pchunk == NO_ENTRY)
		Quit ("CA_CacheMusicChunk: undefined chunk #%d!", chunk);
	    }
	else Quit ("CA_CacheMusicChunk: illegal chunk #%d!", chunk);
	pchunk += STARTMUSIC;

	Diag(CA_DL, "CA_CacheMusicChunk %d (%d)", chunk, pchunk);

	if (audiosegs[pchunk])
		return;

	pos = audiostarts[pchunk];
	length = audiostarts[pchunk+1]-pos;

	ReadSeek(audiohandle, pos, SEEK_SET);

	audiosegs[pchunk] = MM_Alloc(length, "musicchunk");

	ReadBytes(audiohandle, audiosegs[pchunk], length);
}	// CA_CacheMusicChunk

void CA_UnCacheMusicChunk(int chunk)
{	int pchunk;

	// (translate chunk #)

	if (MUS_BEG <= chunk && chunk < MUS_END) {
	    pchunk = MUS_RSM [chunk - MUS_BEG];
	    if (pchunk == NO_ENTRY)
		Quit ("CA_UnCacheMusicChunk: undefined chunk #%d!", chunk);
	    }
	else Quit ("CA_UnCacheMusicChunk: illegal chunk #%d!", chunk);
	pchunk += STARTMUSIC;

	Diag(CA_DL, "CA_UnCacheMusicChunk %d (%d)", chunk, pchunk);

	if (audiosegs[pchunk] == 0) {
		Diag(0, "Trying to free null music chunk %d!", pchunk);
		return;
	}

	MM_Free(audiosegs[pchunk]);
	audiosegs[pchunk] = NULL;
}	// CA_UnCacheMusicChunk

byte *CA_GetMusicChunk (int chunk) {
int pchunk;

// (translate chunk #)

if (MUS_BEG <= chunk && chunk < MUS_END)
	pchunk = MUS_RSM [chunk - MUS_BEG];
else Quit ("CA_GetMusicChunk: illegal chunk #%d!", chunk);

if (pchunk == NO_ENTRY) Quit ("CA_GetMusicChunk: unset chunk #%d!", chunk);

Diag(CA_DL, "CA_GetMusicChunk %d (%d)", chunk, pchunk);

return audiosegs[STARTMUSIC+pchunk];
}	// CA_GetMusicChunk

/* ======================================================================== */

/*
======================
=
= CAL_ExpandGrChunk
=
= Does whatever is needed with a pointer to a compressed chunk
=
======================
*/

void VL_DeModeXize(byte *buf, int width, int height)
{
	byte *mem, *ptr, *destline;
	int plane, x, y;
	
	if (width & 3) {
		Diag(0, "VL_DeModeXize: Not divisible by 4?");
		return;
	}

	mem = MM_Alloc(width * height, 0);

	ptr = buf;

	for (plane = 0; plane != 4; plane++) {
		destline = mem + plane;
		for (y = 0; y != height; y++) {
			for (x = 0; x != width; x += 4)
				destline[x] = *ptr++;
			destline += width;
		}
	}

	memcpy(buf, mem, width * height);

	MM_Free(mem);
}

// NOTE: 'chunk' is already translated
static void CAL_ExpandGrChunk(int chunk, byte *src) {
	int i;
	long expanded;
	int width, height, tilecount;

	if (chunk >= STARTPICS-GFX_BEG && chunk < GFX_RSM [STARTEXTERNS-GFX_BEG]) {
	    if (chunk >= GFX_RSM [TILE8-GFX_BEG]) {
		/* expanded sizes of tile8 are implicit */
		expanded = 8*8*NUMTILE8;
		width = height = 8;
		tilecount = NUMTILE8;
	    } else {
		pictabletype *entry = &pictable [chunk - (STARTPICS-GFX_BEG)];
		width = entry->width;
		height = entry->height;
		expanded = src[0]|(src[1]<<8)|(src[2]<<16)|(src[3]<<24);
		tilecount = 0;
		src += 4;
	    }
	}
	else {
	/* everything else has an explicit size longword */
		width = height = 0;
		tilecount = 0;
		expanded = src[0]|(src[1]<<8)|(src[2]<<16)|(src[3]<<24);
		src += 4;
	}

Diag(CA_DL, "Expanding: %d [%u] (%d*%d) #%d",
	chunk, expanded, width, height, tilecount);

/* allocate final space and decompress it */
	grsegs[chunk] = MM_Alloc(expanded, "grchunk");
	CAL_HuffExpand(src, grsegs[chunk], expanded, grhuffman);

	if (width && height) {
	if (tilecount) {
	    for (i = 0; i < tilecount; i++)
		VL_DeModeXize(grsegs[chunk]+(width*height)*i, width, height);
	} else
		VL_DeModeXize(grsegs[chunk], width, height);
	}
}

/*
======================
=
= CA_CacheGrChunk
=
= Makes sure a given chunk is in memory, loading it if needed
=
======================
*/

void CA_CacheGrChunk(int chunk)
{
	long pos, compressed;
	byte *source;

	int pchunk;

	if (grhandle == -1)
		return;

	// (translate chunk #)

	if (GFX_BEG <= chunk && chunk < GFX_END) {
	    pchunk = GFX_RSM [chunk - GFX_BEG];
	    if (pchunk == NO_ENTRY)
		Quit ("CA_CacheGrChunk: undefined chunk #%d!", chunk);
	    }
	else Quit ("CA_CacheGrChunk: illegal chunk #%d!", chunk);

	Diag(CA_DL, "CacheGrChunk %d (%d)", chunk, pchunk);

	if (grsegs[pchunk])
		return;

/* load the chunk into a buffer */
	pos = grstarts[pchunk];

	compressed = grstarts[pchunk+1]-pos;

	ReadSeek(grhandle, pos, SEEK_SET);

	source = MM_Alloc(compressed, 0);
	ReadBytes(grhandle, source, compressed);

	CAL_ExpandGrChunk(pchunk, source);
	
	MM_Free(source);
}	// CA_CacheGrChunk

void CA_UnCacheGrChunk(int chunk)
{	int pchunk;

	// (translate chunk #)

	if (GFX_BEG <= chunk && chunk < GFX_END)
		pchunk = GFX_RSM [chunk - GFX_BEG];
	else Quit ("CA_UnCacheGrChunk: illegal chunk #%d!", chunk);

	if (pchunk == NO_ENTRY)
		Quit ("CA_UnCacheGrChunk: unset chunk #%d!", chunk);

	Diag(CA_DL, "UnCacheGrChunk %d (%d)", chunk, pchunk);

	if (grsegs[pchunk] == 0) {
		Diag(0, "Trying to free null graphic chunk %d!", pchunk);
		return;
	}

	MM_Free(grsegs[pchunk]);
	grsegs[pchunk] = NULL;
}	// CA_UnCacheGrChunk

byte *CA_GetGrChunk(int chunk) {
int pchunk;

// (translate chunk #)

if (GFX_BEG <= chunk && chunk < GFX_END)
	pchunk = GFX_RSM [chunk - GFX_BEG];
else Quit ("CA_GetGrChunk: illegal chunk #%d!", chunk);

if (pchunk == NO_ENTRY) Quit ("CA_GetGrChunk: unset chunk #%d!", chunk);

Diag(CA_DL, "GetGrChunk %d (%d)", chunk, pchunk);

if (pchunk < NUMGRCHUNKS) return grsegs[pchunk];

Quit("CA_GetGrChunk: unknown pchunk #%d", pchunk);
return 0;
}	// CA_GetGrChunk

////////////////////////////////////////////////////////////////////
//
// Un/Cache a LUMP of graphics
//
////////////////////////////////////////////////////////////////////

// TODO: reverse map of graphic chunks?

void CacheGrLump(int lumpstart, int lumpend) {

// 'lumpstart..lumpend' is logical,
// but real lump ordering is physical. So:

int p_start = GFX_RSM [lumpstart - GFX_BEG];
int p_end = GFX_RSM [lumpend - GFX_BEG];
int i;

for (i = GFX_BEG; i != GFX_END; ++ i) {
	int p_i = GFX_RSM [i - GFX_BEG];

	if (p_i != NO_ENTRY && p_start <= p_i && p_i <= p_end)
		CA_CacheGrChunk (i);
	}
}	// CacheGrLump

void UnCacheGrLump(int lumpstart, int lumpend) {

// 'lumpstart..lumpend' is logical,
// but real lump ordering is physical. So:

int p_start = GFX_RSM [lumpstart - GFX_BEG];
int p_end = GFX_RSM [lumpend - GFX_BEG];
int i;

for (i = GFX_BEG; i != GFX_END; ++ i) {
	int p_i = GFX_RSM [i - GFX_BEG];
	if (p_i != NO_ENTRY && p_start <= p_i && p_i <= p_end)
		CA_UnCacheGrChunk (i);
	}
}	// UnCacheGrLump

// Translate physical graphic # to logical one
// (needed in 'wl_text')
int GetGrLogical (int p_chunk) {
int i;

for (i = GFX_BEG; i != GFX_END; ++ i)
    if (GFX_RSM [i - GFX_BEG] == p_chunk)
	break;

return i;
}	// GetGrLogical

/* ======================================================================== */

/*
======================
=
= CA_CacheMap
=
======================
*/

void CA_CacheMap(int mapnum)
{
	long pos,compressed;
	int plane;
	byte *source;
	word *buffer2seg;
	long expanded;
	
/* load the planes into the already allocated buffers */

	for (plane = 0; plane < MAPPLANES; plane++)
	{
		pos = mapheaderseg[mapnum]->planestart[plane];
		compressed = mapheaderseg[mapnum]->planelength[plane];

		ReadSeek(maphandle, pos, SEEK_SET);
		
		source = MM_Alloc(compressed, 0);

		ReadBytes(maphandle, (byte *)source, compressed);
		
		expanded = source[0] | (source[1] << 8);		
		buffer2seg = MM_Alloc(expanded, "mapdata");

/* NOTE: CarmackExpand implicitly fixes endianness, a RLEW'd only map
         would (likely) need to be swapped in CA_RLEWexpand
         
         Wolfenstein 3D/Spear of Destiny maps are always Carmack'd so this
         case is OK.  CA_RLEWexpand would need to be adjusted for Blake Stone
         and the like.
*/
		CAL_CarmackExpand(source+2, buffer2seg, expanded);
		MM_Free(source);

		expanded = MAPHGT*MAPWDT*2;
		CA_RLEWexpand(buffer2seg+1, mapsegs[plane], expanded, RLEWtag);
		MM_Free(buffer2seg);
	}
}

//
// Digital SFX management
//

static word *DigiList;
static word DigiCount;

static void AudioInit ()
{
        word    *p, *list, pg;
        int     i;

        list = p = (word *) PM_GetPage(PMSoundInfo);
        pg = PMSoundStart;

        for (i = 0; i < PMPageSize / (sizeof(word) * 2); i++, p += 2)
        {
                if (pg >= ChunksInFile - 1)
                        break;
                pg += PMPageNo (p[1]);
        }

        DigiList = MM_Alloc(i * sizeof(word) * 2, "digilist");
        memcpy((void *)DigiList, (void *)list, i * sizeof(word) * 2);

	DigiCount = i;
}	// AudioInit

// TODO: move it to debug
void AudioDump () {
int i;

Diag(0, "Audio info dump:");
for (i = 0; i != DigiCount; ++ i)
	Diag (0, "%u: %u (%u) [%u]", i,
		DigiList[i*2], DigiList[i*2+1], PMPageNo (DigiList[i*2+1]));
}	// AudioDump

// Get overall information about sounds
void GetAudioTotals (unsigned *p_count, unsigned *p_npages) {
if (p_count) *p_count = DigiCount;
if (p_npages) *p_npages = PMSoundInfo - PMSoundStart;
}	// GetAudioTotals

// Get information about `sound`
void GetAudioInfo (unsigned sound, unsigned *p_page, unsigned *p_length) {
if (sound < DigiCount) {
	if (p_page)	*p_page = PMSoundStart + DigiList[sound*2];
	if (p_length)	*p_length = DigiList[1+sound*2];
	}
}	// GetAudioInfo

// Locate sound `page`
int LocateSoundPage (unsigned page) {
if (page >= PMSoundStart) {
    int i;
    page -= PMSoundStart;
    for (i = 0; i != DigiCount; ++ i) {
	unsigned np = PMPageNo (DigiList [i*2+1]);
	if (page < np) return i;
	else page -= np;
	}
    }

return -1;			// (no such page)
}	// LocateSoundPage

/* ======================================================================== */

static unsigned MM_count;

void MM_Startup()
{
MM_count = 0;
}

void MM_Shutdown()
{
if (MM_count)
	Diag(0, "(Memory fragments remaining: %u)", MM_count);
}

void *MM_Alloc(unsigned long size, char *tag) {
// TODO: critical errors handling?

++ MM_count;
return malloc (size);
}		// MM_Alloc

void MM_Free(void *block) {
if (block) {
-- MM_count;

free (block);
}
}		// MM_Free

//
// Sprite management
//

void DeCompileSprite(byte *ptr, byte *buf, BBox *area)
{
	byte *cmdptr;
	byte *pixels;
	int yoff;
	int y, y0, y1;
	int x, left, right;
	int cmd;

	unsigned bt, bb, bl, br;
	if (! ptr) return;

	bt = GFXHGT;
	bb = 0;

	bl = left = ptr[0];
	br = (right = ptr[2]) + 1;

	cmdptr = &ptr[4];

	for (x = left; x <= right; x ++) {
		byte *bp1 = buf + x*GFXWDT;
		cmd = cmdptr[0] | (cmdptr[1] << 8);
		cmdptr += 2;

		while (ptr[cmd]) {
			y1 = ptr[cmd] / 2;
			yoff = (int16_t)(ptr[cmd+2] | (ptr[cmd+3] << 8));
			y0 = ptr[cmd+4] / 2;

			if (y0 < bt) bt = y0;
			if (y1 > bb) bb = y1;
			
			pixels = &ptr[y0 + yoff];
			bp1 += y0;

			for (y = y0; y < y1; y++)
			    *bp1++ = *pixels++;

			bp1 -= y1;
			cmd += 6;
			}
		}	// (x loop)
if (area) {
	area->t = bt, area->b = bb;
	area->l = bl, area->r = br;
	}
}

static byte *sprite_gfx[SPR_TOTAL];
static BBox sprite_box [SPR_TOTAL];

extern int sprite_map [];

byte *GetSprite (int shapenum, BBox **p_box) {
byte *data = sprite_gfx[shapenum];
if (data == NULL) {
	data = MM_Alloc(GFXHGT*GFXWDT, "sprite");
	memset(data, 255, GFXHGT*GFXWDT);
	sprite_gfx[shapenum] = data;

	if (sprite_map[shapenum] == -1)
		Diag(1, "Illegal sprite page #%u", shapenum);

  unsigned page = PMSpriteStart+sprite_map[shapenum];
	DeCompileSprite(PM_GetPage(page), data, sprite_box + shapenum);

	PM_FreePage(page);
	}

if (p_box) *p_box = sprite_box + shapenum;
return data;
}	// GetSprite

void FreeSprites() {
unsigned i;
for (i = 0; i != SPR_TOTAL; ++ i)
		MM_Free (sprite_gfx[i]);
}	// FreeSprites

/*
 * Page management
 */

static boolean PMStarted;

static int PageFile = -1;
int ChunksInFile, PMSpriteStart, PMSoundStart, PMSoundInfo;

PageListStruct *PMPages;

static void PML_ReadFromFile(byte *buf, long offset, word length)
{
	if (!buf)
		Quit("PML_ReadFromFile: Null pointer");
	if (!offset)
		Quit("PML_ReadFromFile: Zero offset");
	if (ReadSeek(PageFile, offset, SEEK_SET) != offset)
		Quit("PML_ReadFromFile: Seek failed");
	if (ReadBytes(PageFile, buf, length) != length)
		Quit("PML_ReadFromFile: Read failed");
}

static void PML_OpenPageFile()
{
	int i;

	PageFile = TryOpenRead(resource_name(pfilename, true));

	/* Read in header variables */
	ChunksInFile = ReadInt16(PageFile);
	PMSpriteStart = ReadInt16(PageFile);
	PMSoundStart = ReadInt16(PageFile);
	PMSoundInfo = ChunksInFile - 1;

	/* Allocate and clear the page list */
	PMPages = MM_Alloc(sizeof(PageListStruct) * ChunksInFile, "pagetab");

	memset(PMPages, 0, sizeof(PageListStruct) * ChunksInFile);

	/* Read in the chunk offsets */
	for (i = 0; i != ChunksInFile; i++)
		PMPages[i].offset = ReadInt32(PageFile);

	/* Read in the chunk lengths */
	for (i = 0; i != ChunksInFile; i++)
		PMPages[i].length = ReadInt16(PageFile);
}

static void PML_ClosePageFile()
{
	if (PageFile != -1)
		CloseRead(PageFile);
		
	if (PMPages) {
		MM_Free(PMPages);
		PMPages = 0;
		}
}

byte *PM_GetPage(int pagenum)
{
	PageListStruct *page;

	if (pagenum >= ChunksInFile)
		Quit("PM_GetPage: Invalid page request");

	page = &PMPages[pagenum];
	if (page->addr == NULL && page->offset != 0) {
		page->addr = MM_Alloc(PMPageSize, "page");
		PML_ReadFromFile(page->addr, page->offset, page->length);
	}
	return page->addr;
}

void PM_FreePage(int pagenum)
{
	PageListStruct *page;
	
	if (pagenum >= ChunksInFile)
		Quit("PM_FreePage: Invalid page request (%u)", pagenum);

	page = &PMPages[pagenum];
	if (page->addr != NULL) {
		MM_Free(page->addr);
		page->addr = NULL;
	}
}
	
void PM_Startup()
{
	if (PMStarted)
		return;

	PML_OpenPageFile();

	PMStarted = true;

	AudioInit ();

Diag(0, "Page file init: total %u pages (%u tiles, %u sprites, %u digifrags, 1 digitab)",
		ChunksInFile,
		PMSpriteStart,
		PMSoundStart - PMSpriteStart,
		PMSoundInfo - PMSoundStart);
}

void PM_Shutdown()
{
	if (!PMStarted)
		return;

  FreeSprites();

  MM_Free(DigiList);

  unsigned pg;
  for (pg = 0; pg != ChunksInFile; ++ pg)
		PM_FreePage(pg);

	PML_ClosePageFile();

Diag(0, "Page file shutdown OK");
}


