
/*
 * +==========================================================
 * |	'id_vh.c':
 * |	High-level video primitives
 * |
 * |	LoneWolf (Wolf3D/Spear of Destiny engine) source code
 * |    by D.G.Gayev (dgaev@mail.ru), 2004-2005
 * +==========================================================
 *
 */

#include "wl_def.h"

int px, py;
byte fontcolor, backcolor;
int fontnumber;

boolean	screenfaded;

static byte palette1[256][3], palette2[256][3];

static unsigned xfrac, yfrac;

#define	XScale(x)	(((x) * xfrac) >> 16)
#define	YScale(y)	(((y) * yfrac) >> 16)

/* ======================================================================== */


// Define this to skip palette fadein/fadeout
#undef SKIPFADE

void VL_FillPalette(int red, int green, int blue)
{
	byte pal[768];
	int i;
	
	for (i = 0; i != sizeof(pal); i += 3) {
		pal[i] = red;
		pal[i+1] = green;
		pal[i+2] = blue;
	}
	VL_SetPalette(pal);
}

/*
=================
=
= VL_FadeOut
=
= Fades the current palette to the given color in the given number of steps
=
=================
*/

void VL_FadeOut(int start, int end, int red, int green, int blue, int steps)
{
#ifdef SKIPFADE

	VL_FillPalette(red, green, blue);

#else

	int i,j,orig;
	byte *origptr, *newptr;

	VL_GetPalette(&palette1[0][0]);
	memcpy(palette2, palette1, 768);

/* fade through intermediate frames */
	for (i = 0; i < steps; i++)
	{
		origptr = &palette1[start][0];
		newptr = &palette2[start][0];
		for (j=start;j<=end;j++)
		{
			orig = *origptr++;
			*newptr++ = orig + (red-orig) * i / steps;
			orig = *origptr++;
			*newptr++ = orig + (green-orig) * i / steps;
			orig = *origptr++;
			*newptr++ = orig + (blue-orig) * i / steps;
		}

		VL_SetPalette(&palette2[0][0]);
	}

/* final color */
	VL_FillPalette(red, green, blue);

#endif

	screenfaded = true;
}

/*
=================
=
= VL_FadeIn
=
=================
*/

void VL_FadeIn(int start, int end, const byte *palette, int steps)
{
#ifdef SKIPFADE

	VL_SetPalette(palette);

#else

	int i, j, delta;

	VL_GetPalette(&palette1[0][0]);
	memcpy(&palette2[0][0],&palette1[0][0],sizeof(palette1));

	start *= 3;
	end = end*3+2;

/* fade through intermediate frames */
	for (i = 0; i < steps; i++)
	{
		for (j = start;j <= end; j++)
		{
			delta = palette[j]-palette1[0][j];
			palette2[0][j] = palette1[0][j] + delta * i / steps;
		}

		VL_SetPalette(&palette2[0][0]);
	}

/* final color */
	VL_SetPalette(palette);

#endif
	screenfaded = false;
}

/*
=================
=
= VL_Bar
=
=================
*/

void VL_Bar(int x, int y, int width, int height, int color)
{
	byte *ptr = VL_BeginDraw() + vwidth * y + x;
	while (height--) {
		memset(ptr, color, width);
		ptr += vwidth;
	}

VL_EndDraw();
}

/*
=================
=
= VL_Plot
=
=================
*/

void VL_Plot(int x, int y, int color)
{
	VL_Bar(XScale(x), YScale(y), XScale(1), YScale(1), color);
}

/*
=================
=
= VL_Hlin
=
=================
*/

void VL_Hlin(int x, int y, int width, int color)
{
	VL_Bar(XScale(x), YScale(y), XScale(width), YScale(1), color);
}

/*
=================
=
= VL_Vlin
=
=================
*/

void VL_Vlin(int x, int y, int height, int color)
{
	VL_Bar(XScale(x), YScale(y), XScale(1), YScale(height), color);
}

/*
=================
=
= VW_Bar
=
=================
*/

void VW_Bar(int x, int y, int width, int height, int color)
{
	VL_Bar(XScale(x), YScale(y), XScale(width), YScale(height), color);
}

/* 
	Font is:
	height is a word
	256 word offsets from start
	256 byte widths
	data
*/

void VW_DrawPropString(char *string)
{
	byte *font;
	int width, step, height, x, xs, y;
	byte *source, *ptrs;
	byte ch;

	font = CA_GetGrChunk(STARTFONT+fontnumber);
	height = font[0] | (font[1] << 8);

	xs = 0;
	
	while ((ch = *string++) != 0) {
		width = step = font[2 + 512 + ch];
		source = font+font[2+ch*2+0]+(font[2+ch*2+1]<<8);
		for (x = 0; x != width; x++) {
			ptrs = source;
			for (y = 0; y != height; y++) {
				if (*ptrs)
					VL_Plot(px+xs, py+y, fontcolor);
				ptrs += step;
			}
			xs++;
			source++;
		}
	}
}

void VW_MeasurePropString(char *string, word *width, word *height)
{
	int w, mw;
	byte *font = CA_GetGrChunk(STARTFONT+fontnumber), ch;
	
	w = 0;
	mw = 0;
	*height = font[0] | (font[1] << 8);
	while ((ch = *string++)) {
		if (ch == '\n') {
			if (mw < w)
				mw = w;

			w = 0;
			*height += font[0] | (font[1] << 8);
		} else {
			w += font[2 + 512 + ch]; 
		}
	}
	if (mw < w)
		mw = w;
		
	*width = mw;
}

/*
=================
=
= VL_MemToScreen
=
= Draws a block of data to the screen.
=
=================
*/

void VL_MemToScreen(const byte *source, int width, int height, int x, int y)
{
unsigned xf = XScale(1), yf = YScale(1), ws = XScale(width);
unsigned w, f;
byte *ptr = VL_BeginDraw() + vwidth * YScale(y) + YScale(x), *buffer;

if (! (buffer = MM_Alloc(ws, 0))) return;

while (height --) {
    byte *bp = buffer;
    for (w = width; w --;) {
	byte val = *source ++;
	for (f = xf; f --;)
	    *bp ++ = val;
	}	// (w loop)
    for (f = yf; f --;) {
	memcpy (ptr, buffer, ws);
	ptr += vwidth;
	}
    }

MM_Free (buffer);
VL_EndDraw();
}	// VL_MemToScreen

// Scale screen coordinates
// (from logical to physical)
void VL_Log2Scr (unsigned *cx, unsigned *cy) {
if (cx) *cx = XScale (*cx);
if (cy) *cy = YScale (*cy);
}	// VL_Log2Scr

void VL_CacheScreen(int chunk)
{
	CA_CacheGrChunk(chunk);
	VL_MemToScreen(CA_GetGrChunk(chunk), ScrWdt, ScrHgt, 0, 0);
	CA_UnCacheGrChunk(chunk);
}

void VWB_DrawTile8(int x, int y, int tile)
{
	VL_MemToScreen(CA_GetGrChunk(TILE8)+(tile*64), 8, 8, x, y);
}

static void VWB_DrawPicF(int x, int y, int chunknum, boolean sync)
{
	int width, height;
	void *source = CA_GetGrChunk(chunknum);

	GetPicDims (chunknum, &width, &height);

	if (source)
		VL_MemToScreen(source, width, height, x, y);
	else		// (clear area)
		VW_Bar (x, y, width, height, 0);

  if (sync) VW_UpdateScreenArea (y, height, x, width);
}

void VWB_DrawPic (int x, int y, int chunknum) {
VWB_DrawPicF (x, y, chunknum, false);
}

void VWB_DrawPicSync (int x, int y, int chunknum) {
VWB_DrawPicF (x, y, chunknum, true);
}

void VW_UpdateScreen() { VL_UpdateScreen(); }

void VW_UpdateScreenArea (unsigned t, unsigned h, unsigned l, unsigned w) {
VL_UpdateScreenArea (YScale(t), YScale(h), XScale(l), XScale(w));
}	// VL_UpdateScreenArea

void VW_Startup()
{
	VL_Startup();
	
	xfrac = (vwidth << 16) / ScrWdt;
	yfrac = (vheight << 16) / ScrHgt;
}

void VW_Shutdown()
{
	VL_Shutdown();
}

